/**********************************************************************
audio play & record (Linux, Solaris, IRIX)

$Id: playrec.c,v 1.2 1999/01/07 11:52:43 purnhage Exp $

Header file: playrec.h

Authors:
HP    Heiko Purnhagen, Uni Hannover <purnhage@tnt.uni-hannover.de>
NM    Nikolaus Meine, Uni Hannover (c/o Heiko Purnhagen)
BE    Bernd Edler, Uni Hannover <edler@tnt.uni-hannover.de>

Changes:
16-sep-98   HP/NM   born, based on auplay.c and aurec.c
            HP/BE   ... and play_list.c
17-sep-98   HP      Solaris & IRIX
07-jan-99   HP      warning if no audio device support
**********************************************************************/

/* Audio play & record for Solaris, Linux, IRIX. */
/* Sample format: 16 bit twos complement, uniform quantisation */

/* Multi channel data is interleaved: l0 r0 l1 r1 ... */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "playrec.h"
#include "common.h"

#ifdef Linux
#include <sys/soundcard.h>
#include <sys/ioctl.h>
#endif

#ifdef Solaris
#include <sys/audioio.h>
#include <unistd.h>
#endif

#ifdef IRIX
#include <dmedia/audio.h>
#endif


/* ---------- declarations ---------- */

#define min(a,b) ((a) < (b) ? (a) : (b))
#define max(a,b) ((a) > (b) ? (a) : (b))

#define IRIXSIZE (1.0)		/* IRIX audio buffer size (sec) */


/* ---------- declarations (structures) ---------- */

struct PRStreamStruct		/* play/rec stream handle */
{
#ifdef Linux
  FILE *f;
#endif

#ifdef Solaris
  FILE *f;
#endif

#ifdef IRIX
  ALconfig rt_config;
  ALport rt_audio_p;
#endif

  long currentSample;		/* number of samples read/written */
  int write;			/* write (not read) flag */
};


/* ---------- variables ---------- */

static int PRdebugLevel = 0;	/* debug level */


/* ---------- functions ---------- */


/* PRInit() */
/* Init audio play & record. */

void PRInit (
  int debugLevel)		/* in: debug level */
				/*     0=off  1=basic  2=full */
{
  PRdebugLevel = debugLevel;
  if (PRdebugLevel >= 1) {
    printf("PRInit: debugLevel=%d\n",PRdebugLevel);
  }
}


/* PROpenRead() */
/* Open play/rec stream for reading. */

PRStream *PROpenRead (
  int device,			/* in: device (0=default) */
  int *numChannel,		/* in/out: number of channels */
  float *fSample)		/* in/out: sampling frequency [Hz] */
				/*  numChannel & fSample are set */
				/*  according to audio device capabilities */
				/* returns: */
				/*  play/rec stream (handle) */
				/*  or NULL if error */
{
  PRStream *s;

#ifdef Linux
  int err,tmp;
  int fd;
#endif  

#ifdef Solaris
  audio_info_t ainfo;
  int err;
  int fd;
#endif  

#ifdef IRIX
  long sr[2];
  long size;
  int err;
#endif  
  
  if (PRdebugLevel)
    printf("PROpenRead: device=%d numChannel=%d fSample=%.1f\n",
	   device,*numChannel,*fSample);

  if ((s=(PRStream*)malloc(sizeof(PRStream)))==NULL)
    CommonExit(-1,"PROpenRead: Can not allocate memory");
  s->currentSample = 0;
  s->write = 0;

#ifdef Linux
  if (!(s->f=fopen("/dev/dsp","r"))) {
    CommonWarning("PROpenRead: Can not open \"%s\"","/dev/dsp");
    free(s);
    return NULL;
  }

  fd = fileno(s->f);
  err = ioctl(fd, SNDCTL_DSP_RESET, 0);
  err = err || ioctl(fd, SNDCTL_DSP_SYNC, NULL);
  if (err) {
    CommonWarning("PROpenRead: Unable to reset/sync dsp");
    free(s);
    return NULL;
  }
  tmp = 16;
  err = err || ioctl(fd, SNDCTL_DSP_SAMPLESIZE, &tmp);
  if (tmp != 16) {
    CommonWarning("PROpenRead: 16 bit audio not supported");
    free(s);
    return NULL;
  }
  tmp = (*numChannel>1)?1:0;
  err = err || ioctl(fd, SNDCTL_DSP_STEREO, &tmp);
  *numChannel = tmp+1;
  tmp = (int)(*fSample+.5);
  err = err || ioctl (fd, SNDCTL_DSP_SPEED, &tmp);
  *fSample = tmp;
  if (err) {
    CommonWarning("PROpenRead: Unable to init dsp");
    free(s);
    return NULL;
  }
#endif

#ifdef Solaris
  if (!(s->f=fopen("/dev/audio","w"))) {
    CommonWarning("PROpenRead: Can not open \"%s\"","/dev/audio");
    free(s);
    return NULL;
  }

  fd = fileno(s->f);

  AUDIO_INITINFO(&ainfo);

  ainfo.record.sample_rate = (long)(*fSample+.5);
  ainfo.record.channels = *numChannel;
  ainfo.record.precision	= 16;
  ainfo.record.encoding	= AUDIO_ENCODING_LINEAR;
  
  err = ioctl(fd,AUDIO_SETINFO,&ainfo);
  err = err || ioctl(fd,AUDIO_GETINFO,&ainfo);
  *fSample = ainfo.record.sample_rate;
  *numChannel = ainfo.record.channels;
  if (err) {
    CommonWarning("PROpenRead: Unable to init audio");
    free(s);
    return NULL;
  }
  if (ainfo.record.precision != 16 ||
      ainfo.record.encoding != AUDIO_ENCODING_LINEAR) {
    CommonWarning("PROpenRead: 16 bit audio not supported");
    free(s);
    return NULL;
  }
#endif

#ifdef IRIX
  sr[0]=AL_INPUT_RATE;
  sr[1]=(long)(*fSample+.5);
  err = 0;
  if (*fSample>0)
    err = err || ALsetparams(AL_DEFAULT_DEVICE,sr,2);
  err = err || ALgetparams(AL_DEFAULT_DEVICE,sr,2);
  *fSample = sr[1];
  s->rt_config = ALnewconfig();
  err = err || (!s->rt_config);
  err = err || ALsetsampfmt(s->rt_config,AL_SAMPFMT_TWOSCOMP);
  err = err || ALsetwidth(s->rt_config,AL_SAMPLE_16); 
  if (*numChannel>0)
    err = err || ALsetchannels(s->rt_config,*numChannel);
  err = err || ((*numChannel=ALgetchannels(s->rt_config))<0);
  size = (long)(*numChannel**fSample*IRIXSIZE);
  err = err || ALsetqueuesize(s->rt_config,size);
  if (err || !(s->rt_audio_p = ALopenport("adc","r",s->rt_config))) {
    CommonWarning("PROpenRead: Unable to init adc");
    free(s);
    return NULL;
  }
#endif

#if !(defined Linux || defined Solaris || defined IRIX)
    CommonWarning("PROpenRead: no audio device support");
    free(s);
    return NULL;
#endif

  if (PRdebugLevel)
    printf("PROpenRead: set to: numChannel=%d fSample=%.1f\n",
	   *numChannel,*fSample);

  return s;
}


/* PROpenWrite() */
/* Open play/rec stream for writing. */

PRStream *PROpenWrite (
  int device,			/* in: device (0=default) */
  int *numChannel,		/* in/out: number of channels */
  float *fSample)		/* in/out: sampling frequency [Hz] */
				/*  numChannel & fSample are set */
				/*  according to audio device capabilities */
				/* returns: */
				/*  play/rec stream (handle) */
				/*  or NULL if error */
{
  PRStream *s;

#ifdef Linux
  int err,tmp;
  int fd;
#endif  

#ifdef Solaris
  audio_info_t ainfo;
  int err;
  int fd;
#endif  

#ifdef IRIX
  long sr[2];
  long size;
  int err;
#endif  
  
  if (PRdebugLevel)
    printf("PROpenWrite: device=%d numChannel=%d fSample=%.1f\n",
	   device,*numChannel,*fSample);

  if ((s=(PRStream*)malloc(sizeof(PRStream)))==NULL)
    CommonExit(-1,"PROpenWrite: Can not allocate memory");
  s->currentSample = 0;
  s->write = 1;

#ifdef Linux
  if (!(s->f=fopen("/dev/dsp","w"))) {
    CommonWarning("PROpenWrite: Can not open \"%s\"","/dev/dsp");
    free(s);
    return NULL;
  }

  fd = fileno(s->f);
  err = ioctl(fd, SNDCTL_DSP_RESET, 0);
  err = err || ioctl(fd, SNDCTL_DSP_SYNC, NULL);
  if (err) {
    CommonWarning("PROpenWrite: Unable to reset/sync dsp");
    free(s);
    return NULL;
  }
  tmp = 16;
  err = err || ioctl(fd, SNDCTL_DSP_SAMPLESIZE, &tmp);
  if (tmp != 16) {
    CommonWarning("PROpenWrite: 16 bit audio not supported");
    free(s);
    return NULL;
  }
  tmp = (*numChannel>1)?1:0;
  err = err || ioctl(fd, SNDCTL_DSP_STEREO, &tmp);
  *numChannel = tmp+1;
  tmp = (int)(*fSample+.5);
  err = err || ioctl (fd, SNDCTL_DSP_SPEED, &tmp);
  *fSample = tmp;
  if (err) {
    CommonWarning("PROpenWrite: Unable to init dsp");
    free(s);
    return NULL;
  }
#endif

#ifdef Solaris
  if (!(s->f=fopen("/dev/audio","w"))) {
    CommonWarning("PROpenWrite: Can not open \"%s\"","/dev/audio");
    free(s);
    return NULL;
  }

  fd = fileno(s->f);

  AUDIO_INITINFO(&ainfo);

  ainfo.play.sample_rate = (long)(*fSample+.5);
  ainfo.play.channels = *numChannel;
  ainfo.play.precision	= 16;
  ainfo.play.encoding	= AUDIO_ENCODING_LINEAR;
  
  err = ioctl(fd,AUDIO_SETINFO,&ainfo);
  err = err || ioctl(fd,AUDIO_GETINFO,&ainfo);
  *fSample = ainfo.play.sample_rate;
  *numChannel = ainfo.play.channels;
  if (err) {
    CommonWarning("PROpenWrite: Unable to init audio");
    free(s);
    return NULL;
  }
  if (ainfo.play.precision != 16 ||
      ainfo.play.encoding != AUDIO_ENCODING_LINEAR) {
    CommonWarning("PROpenWrite: 16 bit audio not supported");
    free(s);
    return NULL;
  }
#endif

#ifdef IRIX
  sr[0]=AL_OUTPUT_RATE;
  sr[1]=(long)(*fSample+.5);
  err = 0;
  if (*fSample>0)
    err = err || ALsetparams(AL_DEFAULT_DEVICE,sr,2);
  err = err || ALgetparams(AL_DEFAULT_DEVICE,sr,2);
  *fSample = sr[1];
  s->rt_config = ALnewconfig();
  err = err || (!s->rt_config);
  err = err || ALsetsampfmt(s->rt_config,AL_SAMPFMT_TWOSCOMP);
  err = err || ALsetwidth(s->rt_config,AL_SAMPLE_16); 
  if (*numChannel>0)
    err = err || ALsetchannels(s->rt_config,*numChannel);
  err = err || ((*numChannel=ALgetchannels(s->rt_config))<0);
  size = (long)(*numChannel**fSample*IRIXSIZE);
  err = err || ALsetqueuesize(s->rt_config,size);
  if (err || !(s->rt_audio_p = ALopenport("dac","w",s->rt_config))) {
    CommonWarning("PROpenWrite: Unable to init dac");
    free(s);
    return NULL;
  }
#endif

#if !(defined Linux || defined Solaris || defined IRIX)
    CommonWarning("PROpenWrite: no audio device support");
    free(s);
    return NULL;
#endif

  if (PRdebugLevel)
    printf("PROpenWrite: set to: numChannel=%d fSample=%.1f\n",
	   *numChannel,*fSample);

  return s;
}


/* PRReadData() */
/* Read data from play/rec stream. */

long PRReadData (
  PRStream *stream,		/* in: play/rec stream (handle) */
  short *data,			/* out: data[] */
  long numSample)		/* in: number of samples to be read */
				/* returns: */
				/*  number of samples read */
{
  long i;

  if (PRdebugLevel > 1)
    printf("PRReadData: numSample=%ld\n",numSample);

  if (stream->write)
    CommonExit(1,"PRReadData: device not in read mode");

#ifdef Linux
  i = fread(data,sizeof(short),numSample,stream->f);
#endif

#ifdef Solaris
  i = fread(data,sizeof(short),numSample,stream->f);
#endif

#ifdef IRIX
  i = (ALreadsamps(stream->rt_audio_p,data,numSample))?0:numSample;
#endif

  stream->currentSample += i;
  return i;
}


/* PRWriteData() */
/* Write data to play/rec stream. */

void PRWriteData (
  PRStream *stream,		/* in: play/rec stream (handle) */
  short *data,			/* in: data[] */
  long numSample)		/* in: number of samples to be written */
{
  if (PRdebugLevel > 1)
    printf("PRWriteData: numSample=%ld\n",numSample);

  if (!stream->write)
    CommonExit(1,"PRWriteData: device not in write mode");

#ifdef Linux
  fwrite(data,sizeof(short),numSample,stream->f);
#endif

#ifdef Solaris
  fwrite(data,sizeof(short),numSample,stream->f);
#endif

#ifdef IRIX
  ALwritesamps(stream->rt_audio_p,data,numSample);
#endif

  stream->currentSample += numSample;
}


/* PRClose() */
/* Close play/rec stream.*/

void PRClose (
  PRStream *stream)		/* in: play/rec stream (handle) */
{
  if (PRdebugLevel)
    printf("PRClose: currentSample=%ld\n",stream->currentSample);

#ifdef Linux
  fclose(stream->f);
#endif

#ifdef Solaris
  fclose(stream->f);
#endif

#ifdef IRIX
  if (stream->write)
    while (ALgetfilled(stream->rt_audio_p) > 0)
      sginap(1);
  ALcloseport(stream->rt_audio_p);
  ALfreeconfig(stream->rt_config);
#endif

  free(stream);
}


/* end of austream.c */

